import copy
import torch
import warnings
from ._prior import Uniform, BinaryPrior, CategoricalPrior, MixedBinaryPrior, MixedCategoricalPrior
from ._prior_update import update_mixed_prior, update_binary_prior, update_categorical_prior, update_continuous_prior
from ._weights import WeightsStabiliser
from ._rchq import recombination
from ._utils import TensorManager


class RecombinationSampler(WeightsStabiliser, TensorManager):
    def __init__(
        self,
        kernel,
        thresh=5,
    ):
        """
        Sampling via kernel recombination.
        
        Args:
        - kernel: class, the class of kernel
        """
        WeightsStabiliser.__init__(self, thresh=thresh)  # WeightsStabiliser class initialisation
        TensorManager.__init__(self) # TensorManager class initialisation
        self.kernel = kernel
        
    def sampling_recombination(
        self,
        X_cand,
        X_nys,
        weights,
        batch_size,
        calc_obj=None,
    ):
        """
        Sampling via kernel recombination.
        
        Args:
        - X_cand: torch.tensor, samples for recombination
        - X_nys: torch.tensor, samples for Nyström approximation
        - weights: torch.tensor, weights
        - batch_size: int, the number of batch samples
        - calc_obj: class, the acquisition function.
        
        Return:
        - idx_rchq: torch.tensor, the indices selected for the next batch
        - w_rchq: torch.tensor, the quadrature weights
        """
        idx_rchq, w_rchq = recombination(
            X_cand,
            X_nys,
            batch_size,
            self.kernel,
            self.device,
            self.dtype,
            init_weights=weights,
            calc_obj=calc_obj,
        )
        return idx_rchq, w_rchq

class EmpiricalSampler(RecombinationSampler):
    def __init__(
        self,
        prior,
        pi,
        kernel,
        thresh=5,
        label="mixedbinary", 
    ):
        """
        Sampling from pi.
        
        Args:
        - prior: class, the class of prior distribution
        - pi: class, the class of pi
        - kernel: class, the class of kernel
        - label: string, prior type. Select from "continuous", "binary", "categorical", "mixedbinary", "mixedcategorical".
        """
        super().__init__(kernel, thresh=thresh)  # RecombinationSampler class initialisation
        #self.prior_initial = copy.deepcopy(prior)
        self.thresh_initial = copy.deepcopy(thresh)
        self.prior = prior
        self.pi = pi
        self.label = label
        self.flag = False
        
    def initialise_prior(self):
        """
        Initialise prior
        """
        if self.label == "continuous":
            self.prior = Uniform(self.prior.bounds)
        elif self.label == "binary":
            self.prior = BinaryPrior(self.prior.n_dims)
        elif self.label == "categorical":
            self.prior = CategoricalPrior(self.prior.categories)
        elif self.label == "mixedbinary":
            self.prior = MixedBinaryPrior(
                self.prior.n_dims_cont, 
                self.prior.n_dims_binary,
                self.prior.bounds,
                self.prior.continous_first,
            )
        elif self.label == "mixedcategorical":
            self.prior = MixedCategoricalPrior(
                self.prior.n_dims_cont, 
                self.prior.n_dims_disc,
                self.prior.categories,
                self.prior.bounds,
                self.prior.continous_first,
            )

    def update_prior(self, X_cand, weights, verbose=False):
        """
        Update prior
        
        Args:
        - X_cand: torch.tensor, samples
        - weights: torch.tensor, weights
        - verbose: bool, show progress if truem otherwise not.
        """
        if self.label == "mixedbinary":
            self.prior = update_mixed_prior(X_cand, weights, self.prior, label="binary")
            if verbose:
                print("The optimised weights")
                print(self.prior.prior_binary.prior_binary.probs)
        elif self.label == "mixedcategorical":
            self.prior = update_mixed_prior(X_cand, weights, self.prior, label="categorical")
            if verbose:
                print("The optimised weights")
                print(self.prior.prior_disc.cat.probs.reshape(
                    self.prior.prior_disc.n_dims,
                    self.prior.prior_disc.n_categories,
                ))
        elif self.label == "continuous":
            self.prior = update_continuous_prior(
                X_cand, weights, self.prior, self.prior.n_dims,
            )
            
        elif self.label == "categorical":
            self.prior = update_categorical_prior(
                weights, X_cand, self.prior,
            )
            if verbose:
                print("The optimised weights")
                print(self.prior.cat.probs.reshape(
                    self.prior.n_dims, self.prior.n_categories
                ))
        elif self.label == "binary":
            self.prior.prior_binary = update_binary_prior(
                weights, X_cand, self.prior.prior_binary,
            )
            if verbose:
                print("The optimised weights")
                print(self.prior.prior_binary.probs)
        else:
            raise ValueError('The domain type should be from "continuous", "binary", "categorical", "mixedbinary", "mixedcategorical"')
    
    def check_categorical(self):
        """
        Check whether or not the domain is categorical.
        
        Return:
        - flag: bool, categorical if true, otherwise not.
        """
        if self.label == "mixedcategorical":
            return True
        elif self.label == "categorical":
            return True
        else:
            return False
    
    def sampling(self, n_rec):
        """
        Sampling from prior with weights
        
        Args:
        - n_rec: int, the number of samples
        
        Return:
        - X_cand: torch.tensor, samples
        - weights: torch.tensor, weights
        """
        X_cand = self.prior.sample(n_rec)
        weights = self.pi(X_cand) / self.prior.pdf(X_cand)
        weights = self.cleansing_weights(weights)
        return X_cand, weights
    
    def categorical_sampling(self, n_rec):
        """
        Sampling from prior with weights
        
        Args:
        - n_rec: int, the number of samples
        
        Return:
        - X_cand: torch.tensor, samples
        - weights: torch.tensor, weights
        """
        X_cand, X_indices = self.prior.sample_both(n_rec)
        weights = self.pi(X_cand) / self.prior.pdf(X_indices)
        weights = self.cleansing_weights(weights)
        return X_cand, X_indices, weights
    
    def recursive_sampling(self, n_rec, n_repeat=5, verbose=False):
        """
        Sampling from prior with weights recursively
        
        Args:
        - n_rec: int, the number of samples
        - n_repeat: int, the number of iterations
        
        Return:
        - X_cand: torch.tensor, samples
        - weights: torch.tensor, weights
        """
        n_accepted = 0
        X_accepted = []
        X_indices_accepted = []
        weights_accepted = []
        self.flag = False
        for i in range(n_repeat):
            if verbose:
                print(str(i)+"-th recursive sampling...")
            if self.check_categorical():
                X_cand, X_indices, weights = self.categorical_sampling(n_rec)
            else:
                X_cand, weights = self.sampling(n_rec)
            
            idx = (weights > 0)
            if not idx.sum() == 0:
                X_accepted.append(X_cand[idx])
                weights_accepted.append(weights[idx])
                n_accepted += idx.sum().item()
                if self.check_categorical():
                    X_indices_accepted.append(X_indices[idx])
            
            if (n_accepted > self.thresh):
                break
        
        if n_accepted == 0:
            if verbose:
                print("Weighted sampling unsuccessful. Uniform random sampling instead...")
            self.flag = True
            if self.check_categorical():
                X_cand, X_indices, weights = self.categorical_sampling(n_rec)
                weights = torch.ones(n_rec) / n_rec
                return X_cand, X_indices, weights
            else:
                X_cand, weights = self.sampling(n_rec)
                weights = torch.ones(n_rec) / n_rec
                return X_cand, weights
        else:
            X_cand = torch.vstack(X_accepted)
            weights = torch.cat(weights_accepted)
            weights = self.cleansing_weights(weights)
            if self.check_categorical():
                X_indices = torch.vstack(X_indices_accepted)
                return X_cand, X_indices, weights
            else:
                return X_cand, weights
    
    def sampling_candidates(self, n_rec, n_nys, verbose=False):        
        """
        Sampling from pi with weights
        
        Args:
        - n_rec: int, the number of samples for recombination
        - n_nys: int, the number of samples for Nyström approximation
        - verbose: bool, show progress if truem otherwise not.
        
        Return:
        - X_cand: torch.tensor, samples for recombination
        - X_nys: torch.tensor, samples for Nyström approximation
        - weights: torch.tensor, weights
        """
        assert n_rec > n_nys
        
        if verbose:
            print("initial sampling...")
        if self.check_categorical():
            X_cand, X_indices, weights = self.categorical_sampling(n_rec)
        else:
            X_cand, weights = self.sampling(n_rec)
        if self.check_weights(weights):
            if verbose:
                print("update prior...")

            if self.check_categorical():
                self.update_prior(X_indices, weights, verbose=verbose)
                self.thresh = n_nys
                X_cand, _, weights = self.recursive_sampling(n_rec, n_repeat=self.thresh, verbose=verbose)
            else:
                self.update_prior(X_cand, weights, verbose=verbose)
                self.thresh = n_nys
                X_cand, weights = self.recursive_sampling(n_rec, n_repeat=self.thresh, verbose=verbose)
        else:
            print("Failed to update prior. Trying recursive sampling...")
            if self.check_categorical():
                X_cand, X_indices, weights = self.recursive_sampling(n_rec, n_repeat=self.thresh, verbose=verbose)
                if self.flag:
                    X_nys = X_cand[:n_nys]
                    return X_cand, X_nys, weights
                self.update_prior(X_indices, weights, verbose=verbose)
                self.thresh = n_nys
                X_cand, _, weights = self.recursive_sampling(n_rec, n_repeat=self.thresh, verbose=verbose)
            else:
                X_cand, weights = self.recursive_sampling(n_rec, n_repeat=self.thresh, verbose=verbose)
                if self.flag:
                    X_nys = X_cand[:n_nys]
                    return X_cand, X_nys, weights
                self.update_prior(X_cand, weights)
                self.thresh = n_nys
                X_cand, weights = self.recursive_sampling(n_rec, n_repeat=self.thresh, verbose=verbose)
        
        if self.label == "continuous":
            X_nys = self.kmeans_resampling(X_cand, n_clusters=n_nys)
        else:
            idx_nys = self.deweighted_resampling(weights, n_nys)
            X_nys = X_cand[idx_nys]
        
        self.thresh = copy.deepcopy(self.thresh_initial)
        return X_cand, X_nys, weights
    
    def adaptive_pruning(self, weights, n_rec, n_nys, thresh=1e-3):
        """
        Pruning the candindates from dataset with weights
        
        Args:
        - weights: torch.tensor, weights
        - n_rec: int, the number of samples for recombination
        - n_nys: int, the number of samples for Nyström approximation
        
        Return:
        - idx_sampled: torch.tensor, indices where X_cand is sampled
        """
        indices = weights.argsort(descending=True)
        try:
            n_accepted = torch.where(weights[indices] > thresh)[0][-1] + 1
            if n_accepted >= n_rec:
                n_pruned = n_rec
            elif n_nys >= n_accepted:
                n_pruned = n_nys
            else:
                n_pruned = n_accepted
        except:
            n_pruned = n_nys
        idx_sampled = indices[:n_pruned]
        return idx_sampled
    
    def sampling_datasets(self, n_rec, n_nys):
        """
        Sampling from dataset with weights
        
        Args:
        - n_rec: int, the number of samples for recombination
        - n_nys: int, the number of samples for Nyström approximation
        
        Return:
        - X_cand: torch.tensor, samples for recombination
        - X_nys: torch.tensor, samples for Nyström approximation
        - weights: torch.tensor, weights
        - idx_sampled: (optional) torch.tensor, indices where X_cand is sampled
        """
        assert n_rec > n_nys
    
        X_cand = self.prior.available_candidates()
        weights = self.pi(X_cand)
        
        if self.dataset_pruning:
            idx_sampled = self.adaptive_pruning(weights, n_rec, n_nys)
            X_cand = X_cand[idx_sampled]
            weights = weights[idx_sampled]
        
        weights = self.cleansing_weights(weights)
        idx_nys = self.deweighted_resampling(weights, n_nys)
        X_nys = X_cand[idx_nys]
        
        if self.dataset_pruning:
            return idx_sampled, X_cand, X_nys, weights
        else:
            return X_cand, X_nys, weights

class MixtureSampler:
    def __init__(self, prior, sober, ratio_wkde=0.5):
        """
        Sampling from posterior.
        
        Args:
        - prior: class, the class of prior distribution
        - sober: class, the class of the learnt Sober
        - ratio_wkde: float, the proportion to sample from pi
        """
        self.prior = prior
        self.sober = sober
        self.bounds = prior.bounds
        self.ratio_wkde = ratio_wkde
    
    def sample(self, n_samples):
        """
        Sampling from the mixture of prior and pi
        
        Args:
           - n_samples: int, number of samples to draw
           
        Returns:
            - samples: torch.tensor, the samples from mixture density
        """
        n_wkde = int(self.ratio_wkde * n_samples)
        n_prior = int((1-self.ratio_wkde) * n_samples)

        tm = TensorManager()
        
        if n_wkde:
            samples_wkde = self.sober.prior.sample(n_wkde)
        else:
            samples_wkde = torch.zeros(
                [0, self.sober.prior.n_dims], dtype=tm.dtype, device=tm.device
            )
        if n_prior:
            samples_prior = self.prior.sample(n_prior)
        else:
            samples_prior = torch.zeros(
                [0, self.sober.prior.n_dims], dtype=tm.dtype, device=tm.device
            )
        samples = torch.vstack([samples_wkde, samples_prior])
        return samples
    
    def pdf(self, X):
        """
        Probability density function of the mixture distribution
        
        Args:
           - n_samples: int, number of samples to draw
           
        Returns:
            - samples: torch.tensor, the samples from mixture density
        """
        pdfs = self.sober.prior.pdf(X)/2 + self.prior.pdf(X)/2
        return pdfs



import copy
import torch
import torch.distributions as D


class RBFHyperPrior:
    def __init__(self, theta_map=None):
        """
        Hyperprior for the RBF kernel hyperparameters of FITBO Gaussian process model.
        Hyperprior is set as log-normal distribution. We set hyperprior on log-transformed space.
        While theta is a set of log-space hyperparmeters, Theta is a set of normal space hyperparamters
        theta = [eta, lik_var, lengthscale, outputscale]
        eta: the derivation from the current maximum of the observation, eta = y - max(Yobs)
        lik_var: the noise variance for the likelihood
        lengthscale: the lengthscale hyperparameter of RBF kernel
        outputscale: the outputscale hyperparameter of RBF kernel
        
        Args:
        - theta_map: torch.tensor or None, type-II MLE estimated hyperparameters.
        """
        self.initialise(theta_map)
        
    def initial_hyperprior(self, theta_map=None):
        """
        Initial hyperparameters recommended by the following paper.
        https://proceedings.neurips.cc/paper/2016/hash/3bbfdde8842a5c44a0323518eec97cbe-Abstract.html
        
        Args:
        - theta_map: torch.tensor or None, type-II MLE estimated hyperparameters.
        
        Return:
        - hypermu: the mean of log-space multivariate normal distribution
        - hypercov: the covariance matrix of log-space multivariate normal distribution
        """
        if theta_map == None:
            hypermu = torch.tensor([-2, 0.1, 0.1, 0.4])
            hyperstd = torch.tensor([0.7, 1, 0.7, 0.7])
        else:
            hypermu = torch.cat([
                torch.tensor([-2]),
                theta_map.log(),
            ])
            hyperstd = torch.tensor([0.1, 0.1, 0.1, 0.1])
        
        hypercov = hyperstd.pow(2).diag()
        return hypermu, hypercov
    
    def initialise(self, theta_map):
        """
        Initialise hyper-hyperparameters of hyperprior
        
        Args:
        - theta_map: torch.tensor or None, type-II MLE estimated hyperparameters.
        """
        self.hypermu, self.hypercov = self.initial_hyperprior(theta_map)
        self.mvn = D.MultivariateNormal(
            self.hypermu, self.hypercov,
        )
    
    def sample(self, n_samples):
        """
        Sampling from hyperprior
        
        Args:
        - n_samples: int
        
        Return:
        - hypersamples: torch.tensor, the hypersamples drawn from hyperprior
        """
        return self.mvn.sample(torch.Size([n_samples]))
    
    def pdf(self, hypersamples):
        """
        Compute the probability density function (PDF) values
        
        Args:
        - hypersamples: torch.tensor, the hypersamples of log-transformed hyperparameters
        
        Return:
        - pdfs: torch.tensor, the PDF values
        """
        return self.mvn.log_prob(hypersamples).exp()



import torch
import warnings
import gpytorch
from botorch.fit import fit_gpytorch_model
from gpytorch.priors.torch_priors import GammaPrior


class ExactGPModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y, likelihood, gp_kernel):
        """
        Args:
            - train_x: torch.tensor, inputs. torch.Size(n_data, n_dims)
            - train_y: torch.tensor, observations
            - likelihood: gpytorch.likelihoods, GP likelihood model
            - gp_kernel: gpytorch.kernels, GP kernel model
        """
        super(ExactGPModel, self).__init__(train_x, train_y, likelihood)
        self.mean_module = gpytorch.means.ZeroMean()
        self.covar_module = gp_kernel

    def forward(self, x):
        """
        Args:
            - x: torch.tensor, inputs. torch.Size(n_data, n_dims)

        Returns:
            - torch.distributions, predictive posterior distribution at given x
        """
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)


def set_gp(train_x, train_y, gp_kernel, device, lik=1e-10, rng=10, train_lik=False):
    """
    We can select whether or not to train likelihood variance.
    The true_likelihood query must be noiseless, so learning GP likelihood noise variance could be redundant.
    However, likelihood noise variance plays an important role in a limited number of samples in the early stage.
    So, setting interval constraints keeps the likelihood noise variance within a safe area.
    Otherwise, GP could confuse the meaningful multimodal peaks of true_likelihood as noise.

    Args:
        - train_x: torch.tensor, inputs. torch.Size(n_data, n_dims)
        - train_y: torch.tensor, observations
        - gp_kernel: gpytorch.kernels, GP kernel model
        - device: torch.device, cpu or cuda
        - lik: float, the initial value of GP likelihood noise variance
        - rng: int, tne range coefficient of GP likelihood noise variance
        - train_like: bool, flag whether or not to update GP likelihood noise variance

    Returns:
        - model: gpytorch.models, function of GP model.
    """
    likelihood = gpytorch.likelihoods.GaussianLikelihood()
    likelihood.noise_covar.register_constraint("raw_noise", gpytorch.constraints.Interval(lik / rng, lik * rng))
    model = ExactGPModel(train_x, train_y, likelihood, gp_kernel)
    model.covar_module.base_kernel.lengthscale_prior = GammaPrior(3.0, 6.0)
    model.covar_module.outputscale_prior = GammaPrior(2.0, 0.15)
    hypers = {
        'likelihood.noise_covar.noise': torch.tensor(lik),
    }

    model.initialize(**hypers)
    if not train_lik:
        model.likelihood.raw_noise.requires_grad = False

    if device.type == 'cuda':
        model = model.cuda()
        model.likelihood = model.likelihood.cuda()
    return model


class Closure:
    """
    Args:
        - mll: gpytorch.mlls.ExactMarginalLogLikelihood, marginal log likelihood
        - optimiser: torch.optim, L-BFGS-B optimizer from FullBatchLBFGS

    Returns:
        - loss: torch.tensor, negative log marginal likelihood of GP
    """
    def __init__(self, mll, optimizer):
        self.mll = mll
        self.optimizer = optimizer
        self.train_inputs, self.train_targets = mll.model.train_inputs, mll.model.train_targets

    def __call__(self):
        self.optimizer.zero_grad()
        with gpytorch.settings.fast_computations(log_prob=True):
            output = self.mll.model(*self.train_inputs)
            args = [output, self.train_targets]
            loss = -self.mll(*args).sum()
        return loss


def train_GP_with_BFGS(mll, training_iter, thresh):
    """
    L-BFGS-B implementation is from https://github.com/hjmshi/PyTorch-LBFGS

    Args:
        - mll: gpytorch.mlls.ExactMarginalLogLikelihood, marginal log likelihood
        - training_iter: int, the maximum number of iteration of optimisation loop
        - thresh: float, the stopping criterion

    Returns:
        - mll: gpytorch.mlls.ExactMarginalLogLikelihood, marginal log likelihood
    """
    # Use full-batch L-BFGS optimizer
    optimizer = FullBatchLBFGS(mll.model.parameters())
    closure = Closure(mll, optimizer)
    loss = closure()
    loss.backward()
    loss_best = torch.tensor(1e10)

    for i in range(training_iter):
        # perform step and update curvature
        options = {'closure': closure, 'current_loss': loss, 'max_ls': 10}
        loss, _, lr, _, F_eval, G_eval, _, _ = optimizer.step(options)

        if loss.item() < loss_best:
            delta = torch.abs(loss_best - loss.detach())
            loss_best = loss.item()
            if delta < thresh:
                break
    return mll


def train_GP_with_Adam(mll, lr, training_iter, thresh):
    """
    Args:
        - mll: gpytorch.mlls.ExactMarginalLogLikelihood, marginal log likelihood
        - lr: float, the learning rate
        - training_iter: int, the maximum number of iteration of optimisation loop
        - thresh: float, the stopping criterion

    Returns:
        - mll: gpytorch.mlls.ExactMarginalLogLikelihood, marginal log likelihood
    """
    optimizer = torch.optim.Adam(mll.model.parameters(), lr=lr)
    train_x = mll.model.train_inputs[0]
    train_y = mll.model.train_targets
    loss_best = torch.tensor(1e10)

    for i in range(training_iter):
        optimizer.zero_grad()
        output = mll.model(train_x)
        loss = -mll(output, train_y)
        loss.backward()
        optimizer.step()
        if loss.item() < loss_best:
            delta = torch.abs(loss_best - loss.detach())
            loss_best = loss.item()
            if delta < thresh:
                break
    return mll


def train_GP(model, training_iter=50, thresh=0.01, lr=0.1, optimiser="BoTorch"):
    """
    Args:
        - model: gpytorch.models, function of GP model.
        - train_iter: int, the maximum iteration for GP hyperparameter training.
        - thresh: float, the threshold as a stopping criterion of GP hyperparameter training.
        - lr: float, the learning rate of Adam optimiser
        - optimiser: string, select the optimiser ["L-BFGS-B", "BoTorch", "Adam"]

    Returns:
        - model: gpytorch.models, function of GP model.
    """
    model.train()
    model.likelihood.train()
    mll = gpytorch.mlls.ExactMarginalLogLikelihood(model.likelihood, model)
    try:
        if optimiser == "BoTorch":
            mll = fit_gpytorch_model(mll)
        elif optimiser == "L-BFGS-B":
            mll = train_GP_with_BFGS(mll, training_iter, thresh)

        elif optimiser == "Adam":
            mll = train_GP_with_Adam(mll, lr, training_iter, thresh)
        else:
            raise Exception("The given optimiser is not defined")
    except:
        warnings.warn("Optimiser " + optimiser + " failed. Optimising again with Adam...")
        mll = train_GP_with_Adam(mll, lr, training_iter, thresh)
    return model


def update_gp(train_x, train_y, gp_kernel, device, lik=1e-10, training_iter=50, thresh=0.01, lr=0.1, rng=10, train_lik=False, optimiser="BoTorch"):
    """
    Input:
        - train_x: torch.tensor, inputs. torch.Size(n_data, n_dims)
        - train_y: torch.tensor, observations
        - gp_kernel: gpytorch.kernels, GP kernel model
        - device: torch.device, cpu or cuda
        - lik: float, the initial value of GP likelihood noise variance
        - train_iter: int, the maximum iteration for GP hyperparameter training.
        - thresh: float, the threshold as a stopping criterion of GP hyperparameter training.
        - lr: float, the learning rate of Adam optimiser
        - rng: int, tne range coefficient of GP likelihood noise variance
        - train_like: bool, flag whether or not to update GP likelihood noise variance
        - optimiser: string, select the optimiser ["L-BFGS-B", "BoTorch", "Adam"]

    Output:
        - model: gpytorch.models, function of GP model.
    """
    model = set_gp(train_x, train_y, gp_kernel, device, lik=lik, rng=rng, train_lik=train_lik)
    model = train_GP(model, training_iter=training_iter, thresh=thresh, lr=lr, optimiser=optimiser)
    return model


def predict(test_x, model):
    """
    Fast variance inference is made with LOVE via fast_pred_var().
    For accurate variance inference, you can just comment out the part.

    Input:
        - model: gpytorch.models, function of GP model.

    Output:
        - pred.mean; torch.tensor, the predictive mean
        - pred.variance; torch.tensor, the predictive variance
    """
    model.eval()
    model.likelihood.eval()

    with torch.no_grad():
        try:
            with gpytorch.settings.fast_pred_var():
                pred = model.likelihood(model(test_x))
        except:
            try:
                pred = model.likelihood(model(test_x))
            except:
                warnings.warn("Cholesky failed. Adding more jitter...")
                with gpytorch.settings.cholesky_jitter(float_value=1e-2):
                    pred = model.likelihood(model(test_x))
    return pred.mean, pred.variance

def predict_mean(test_x, model):
    """
    Fast variance inference is made with LOVE via fast_pred_var().
    For accurate variance inference, you can just comment out the part.

    Input:
        - model: gpytorch.models, function of GP model.

    Output:
        - pred.mean; torch.tensor, the predictive mean
        - pred.variance; torch.tensor, the predictive variance
    """
    pred_mean, _ = predict(test_x, model)
    return pred_mean

def get_cov_cache(model):
    """
    woodbury_inv = K(Xobs, Xobs)^(-1)
    S @ S.T = woodbury_inv

    Input:
        - model: gpytorch.models, function of GP model, typically self.wsabi.model in _basq.py

    Output:
        - woodbury_inv: torch.tensor, the inverse of Gram matrix K(Xobs, Xobs)^(-1)
        - Xobs: torch.tensor, the observed inputs X
        - lik_var: torch.tensor, the GP likelihood noise variance
    """
    Xobs = model.train_inputs[0]
    lik_var = model.likelihood.noise
    try:
        S = model.prediction_strategy.covar_cache
    except:
        model.eval()
        mean = Xobs[0].unsqueeze(0)
        model(mean)
        S = model.prediction_strategy.covar_cache
    woodbury_inv = S @ S.T
    return woodbury_inv, Xobs, lik_var


def predictive_covariance(x, y, model):
    """
    Input:
        - x: torch.tensor, inputs x
        - y: torch.tensor, inputs y
        - model: gpytorch.models, function of GP model.

    Output:
        - cov_xy: torch.tensor, predictive covariance matrix
    """
    woodbury_inv, Xobs, lik_var = get_cov_cache(model)
    Kxy = model.covar_module.forward(x, y)
    KxX = model.covar_module.forward(x, Xobs)
    KXy = model.covar_module.forward(Xobs, y)
    cov_xy = Kxy - KxX @ woodbury_inv @ KXy

    """
    if len(x.shape) == 3 or len(y.shape) == 3:
        d = min(x.shape[1], y.shape[1])
        cov_xy[:, range(d), range(d)] += lik_var        
    else:
        d = min(len(x), len(y))
        cov_xy[range(d), range(d)] += lik_var
    """
    return cov_xy



import numpy as np
import math
import torch
from scipy import special
from scipy import optimize


EPS = 10e-15

class TruncatedMVN:
    """
    Create a normal distribution :math:`X  \sim N ({\mu}, {\Sigma})` subject to linear inequality constraints
    :math:`lb < X < ub` and sample from it using minimax tilting. Based on the MATLAB implemention by the authors
    (reference below).

    :param np.ndarray mu: (size D) mean of the normal distribution :math:`\mathbf {\mu}`.
    :param np.ndarray cov: (size D x D) covariance of the normal distribution :math:`\mathbf {\Sigma}`.
    :param np.ndarray lb: (size D) lower bound constrain of the multivariate normal distribution :math:`\mathbf lb`.
    :param np.ndarray ub: (size D) upper bound constrain of the multivariate normal distribution :math:`\mathbf ub`.
    :param Union[int, None] seed: a random seed.

    Note that the algorithm may not work if 'cov' is close to being rank deficient.

    Reference:
    Botev, Z. I., (2016), The normal law under linear restrictions: simulation and estimation via minimax tilting,
    Journal of the Royal Statistical Society Series B, 79, issue 1, p. 125-148,

    Example:
        >>> d = 10  # dimensions
        >>>
        >>> # random mu and cov
        >>> mu = np.random.rand(d)
        >>> cov = 0.5 - np.random.rand(d ** 2).reshape((d, d))
        >>> cov = np.triu(cov)
        >>> cov += cov.T - np.diag(cov.diagonal())
        >>> cov = np.dot(cov, cov)
        >>>
        >>> # constraints
        >>> lb = np.zeros_like(mu) - 2
        >>> ub = np.ones_like(mu) * np.inf
        >>>
        >>> # create truncated normal and sample from it
        >>> n_samples = 100000
        >>> samples = TruncatedMVN(mu, cov, lb, ub).sample(n_samples)

    Reimplementation by Paul Brunzema
    """

    def __init__(self, mu, cov, bounds, seed=None):
        mu = mu.numpy()
        cov = cov.numpy()
        lb = bounds[0].numpy()
        ub = bounds[1].numpy()
        
        self.dim = len(mu)
        if not cov.shape[0] == cov.shape[1]:
            raise RuntimeError("Covariance matrix must be of shape DxD!")
        if not (self.dim == cov.shape[0] and self.dim == len(lb) and self.dim == len(ub)):
            raise RuntimeError("Dimensions D of mean (mu), covariance matric (cov), lower bound (lb) "
                               "and upper bound (ub) must be the same!")

        self.cov = cov
        self.orig_mu = mu
        self.orig_lb = lb
        self.orig_ub = ub
        
        # permutated
        self.lb = lb - mu  # move distr./bounds to have zero mean
        self.ub = ub - mu  # move distr./bounds to have zero mean
        if np.any(self.ub <= self.lb):
            raise RuntimeError("Upper bound (ub) must be strictly greater than lower bound (lb) for all D dimensions!")

        # scaled Cholesky with zero diagonal, permutated
        self.L = np.empty_like(cov)
        self.unscaled_L = np.empty_like(cov)

        # placeholder for optimization
        self.perm = None
        self.x = None
        self.mu = None
        self.psistar = None

        # for numerics
        self.eps = EPS

        # a random state
        self.random_state = np.random.RandomState(seed)

    def sample(self, n):
        """
        Create n samples from the truncated normal distribution.

        :param int n: Number of samples to create.
        :return: D x n array with the samples.
        :rtype: np.ndarray
        """
        if not isinstance(n, int):
            raise RuntimeError("Number of samples must be an integer!")

        # factors (Cholesky, etc.) only need to be computed once!
        if self.psistar is None:
            self.compute_factors()

        # start acceptance rejection sampling
        rv = np.array([], dtype=np.float64).reshape(self.dim, 0)
        accept, iteration = 0, 0
        while accept < n:
            logpr, Z = self.mvnrnd(n, self.mu)  # simulate n proposals
            idx = -np.log(self.random_state.rand(n)) > (self.psistar - logpr)  # acceptance tests
            rv = np.concatenate((rv, Z[:, idx]), axis=1)  # accumulate accepted
            accept = rv.shape[1]  # keep track of # of accepted
            iteration += 1
            if iteration == 10 ** 3:
                print('Warning: Acceptance prob. smaller than 0.001.')
            elif iteration > 10 ** 4:
                accept = n
                rv = np.concatenate((rv, Z), axis=1)
                print('Warning: Sample is only approximately distributed.')

        # finish sampling and postprocess the samples!
        order = self.perm.argsort(axis=0)
        rv = rv[:, :n]
        rv = self.unscaled_L @ rv
        rv = rv[order, :]

        # retransfer to original mean
        rv += np.tile(self.orig_mu.reshape(self.dim, 1), (1, rv.shape[-1]))  # Z = X + mu
        return torch.from_numpy(rv).float().T
    
    def compute_factors(self):
        # compute permutated Cholesky factor and solve optimization

        # Cholesky decomposition of matrix with permuation
        self.unscaled_L, self.perm = self.colperm()
        D = np.diag(self.unscaled_L)
        if np.any(D < self.eps):
            print('Warning: Method might fail as covariance matrix is singular!')

        # rescale
        scaled_L = self.unscaled_L / np.tile(D.reshape(self.dim, 1), (1, self.dim))
        self.lb = self.lb / D
        self.ub = self.ub / D

        # remove diagonal
        self.L = scaled_L - np.eye(self.dim)

        # get gradient/Jacobian function
        gradpsi = self.get_gradient_function()
        x0 = np.zeros(2 * (self.dim - 1))

        # find optimal tilting parameter non-linear equation solver
        sol = optimize.root(gradpsi, x0, args=(self.L, self.lb, self.ub), method='hybr', jac=True)
        if not sol.success:
            print('Warning: Method may fail as covariance matrix is close to singular!')
        self.x = sol.x[:self.dim - 1]
        self.mu = sol.x[self.dim - 1:]

        # compute psi star
        self.psistar = self.psy(self.x, self.mu)
        
    def reset(self):
        # reset factors -> when sampling, optimization for optimal tilting parameters is performed again

        # permutated
        self.lb = self.orig_lb - self.orig_mu  # move distr./bounds to have zero mean
        self.ub = self.orig_ub - self.orig_mu

        # scaled Cholesky with zero diagonal, permutated
        self.L = np.empty_like(self.cov)
        self.unscaled_L = np.empty_like(self.cov)

        # placeholder for optimization
        self.perm = None
        self.x = None
        self.mu = None
        self.psistar = None

    def mvnrnd(self, n, mu):
        # generates the proposals from the exponentially tilted sequential importance sampling pdf
        # output:     logpr, log-likelihood of sample
        #             Z, random sample
        mu = np.append(mu, [0.])
        Z = np.zeros((self.dim, n))
        logpr = 0
        for k in range(self.dim):
            # compute matrix multiplication L @ Z
            col = self.L[k, :k] @ Z[:k, :]
            # compute limits of truncation
            tl = self.lb[k] - mu[k] - col
            tu = self.ub[k] - mu[k] - col
            # simulate N(mu,1) conditional on [tl,tu]
            Z[k, :] = mu[k] + self.trandn(tl, tu)
            # update likelihood ratio
            logpr += lnNormalProb(tl, tu) + .5 * mu[k] ** 2 - mu[k] * Z[k, :]
        return logpr, Z

    def trandn(self, lb, ub):
        """
        Sample generator for the truncated standard multivariate normal distribution :math:`X \sim N(0,I)` s.t.
        :math:`lb<X<ub`.

        If you wish to simulate a random variable 'Z' from the non-standard Gaussian :math:`N(m,s^2)`
        conditional on :math:`lb<Z<ub`, then first simulate x=TruncatedMVNSampler.trandn((l-m)/s,(u-m)/s) and set
        Z=m+s*x.
        Infinite values for 'ub' and 'lb' are accepted.

        :param np.ndarray lb: (size D) lower bound constrain of the normal distribution :math:`\mathbf lb`.
        :param np.ndarray ub: (size D) upper bound constrain of the normal distribution :math:`\mathbf lb`.

        :return: D samples if the truncated normal distribition x ~ N(0, I) subject to lb < x < ub.
        :rtype: np.ndarray
        """
        if not len(lb) == len(ub):
            raise RuntimeError("Lower bound (lb) and upper bound (ub) must be of the same length!")

        x = np.empty_like(lb)
        a = 0.66  # threshold used in MATLAB implementation
        # three cases to consider
        # case 1: a<lb<ub
        I = lb > a
        if np.any(I):
            tl = lb[I]
            tu = ub[I]
            x[I] = self.ntail(tl, tu)
        # case 2: lb<ub<-a
        J = ub < -a
        if np.any(J):
            tl = -ub[J]
            tu = -lb[J]
            x[J] = - self.ntail(tl, tu)
        # case 3: otherwise use inverse transform or accept-reject
        I = ~(I | J)
        if np.any(I):
            tl = lb[I]
            tu = ub[I]
            x[I] = self.tn(tl, tu)
        return x

    def tn(self, lb, ub, tol=2):
        # samples a column vector of length=len(lb)=len(ub) from the standard multivariate normal distribution
        # truncated over the region [lb,ub], where -a<lb<ub<a for some 'a' and lb and ub are column vectors
        # uses acceptance rejection and inverse-transform method

        sw = tol  # controls switch between methods, threshold can be tuned for maximum speed for each platform
        x = np.empty_like(lb)
        # case 1: abs(ub-lb)>tol, uses accept-reject from randn
        I = abs(ub - lb) > sw
        if np.any(I):
            tl = lb[I]
            tu = ub[I]
            x[I] = self.trnd(tl, tu)

        # case 2: abs(u-l)<tol, uses inverse-transform
        I = ~I
        if np.any(I):
            tl = lb[I]
            tu = ub[I]
            pl = special.erfc(tl / np.sqrt(2)) / 2
            pu = special.erfc(tu / np.sqrt(2)) / 2
            x[I] = np.sqrt(2) * special.erfcinv(2 * (pl - (pl - pu) * self.random_state.rand(len(tl))))
        return x

    def trnd(self, lb, ub):
        # uses acceptance rejection to simulate from truncated normal
        x = self.random_state.randn(len(lb))  # sample normal
        test = (x < lb) | (x > ub)
        I = np.where(test)[0]
        d = len(I)
        while d > 0:  # while there are rejections
            ly = lb[I]
            uy = ub[I]
            y = self.random_state.randn(len(uy))  # resample
            idx = (y > ly) & (y < uy)  # accepted
            x[I[idx]] = y[idx]
            I = I[~idx]
            d = len(I)
        return x

    def ntail(self, lb, ub):
        # samples a column vector of length=len(lb)=len(ub) from the standard multivariate normal distribution
        # truncated over the region [lb,ub], where lb>0 and lb and ub are column vectors
        # uses acceptance-rejection from Rayleigh distr. similar to Marsaglia (1964)
        if not len(lb) == len(ub):
            raise RuntimeError("Lower bound (lb) and upper bound (ub) must be of the same length!")
        c = (lb ** 2) / 2
        n = len(lb)
        f = np.expm1(c - ub ** 2 / 2)
        x = c - np.log(1 + self.random_state.rand(n) * f)  # sample using Rayleigh
        # keep list of rejected
        I = np.where(self.random_state.rand(n) ** 2 * x > c)[0]
        d = len(I)
        while d > 0:  # while there are rejections
            cy = c[I]
            y = cy - np.log(1 + self.random_state.rand(d) * f[I])
            idx = (self.random_state.rand(d) ** 2 * y) < cy  # accepted
            x[I[idx]] = y[idx]  # store the accepted
            I = I[~idx]  # remove accepted from the list
            d = len(I)
        return np.sqrt(2 * x)  # this Rayleigh transform can be delayed till the end

    def psy(self, x, mu):
        # implements psi(x,mu); assumes scaled 'L' without diagonal
        x = np.append(x, [0.])
        mu = np.append(mu, [0.])
        c = self.L @ x
        lt = self.lb - mu - c
        ut = self.ub - mu - c
        p = np.sum(lnNormalProb(lt, ut) + 0.5 * mu ** 2 - x * mu)
        return p

    def get_gradient_function(self):
        # wrapper to avoid dependancy on self

        def gradpsi(y, L, l, u):
            # implements gradient of psi(x) to find optimal exponential twisting, returns also the Jacobian
            # NOTE: assumes scaled 'L' with zero diagonal
            d = len(u)
            c = np.zeros(d)
            mu, x = c.copy(), c.copy()
            x[0:d - 1] = y[0:d - 1]
            mu[0:d - 1] = y[d - 1:]

            # compute now ~l and ~u
            c[1:d] = L[1:d, :] @ x
            lt = l - mu - c
            ut = u - mu - c

            # compute gradients avoiding catastrophic cancellation
            w = lnNormalProb(lt, ut)
            pl = np.exp(-0.5 * lt ** 2 - w) / np.sqrt(2 * math.pi)
            pu = np.exp(-0.5 * ut ** 2 - w) / np.sqrt(2 * math.pi)
            P = pl - pu

            # output the gradient
            dfdx = - mu[0:d - 1] + (P.T @ L[:, 0:d - 1]).T
            dfdm = mu - x + P
            grad = np.concatenate((dfdx, dfdm[:-1]), axis=0)

            # construct jacobian
            lt[np.isinf(lt)] = 0
            ut[np.isinf(ut)] = 0

            dP = - P ** 2 + lt * pl - ut * pu
            DL = np.tile(dP.reshape(d, 1), (1, d)) * L
            mx = DL - np.eye(d)
            xx = L.T @ DL
            mx = mx[:-1, :-1]
            xx = xx[:-1, :-1]
            J = np.block([[xx, mx.T],
                          [mx, np.diag(1 + dP[:-1])]])
            return (grad, J)

        return gradpsi

    def colperm(self):
        perm = np.arange(self.dim)
        L = np.zeros_like(self.cov)
        z = np.zeros_like(self.orig_mu)

        for j in perm.copy():
            pr = np.ones_like(z) * np.inf  # compute marginal prob.
            I = np.arange(j, self.dim)  # search remaining dimensions
            D = np.diag(self.cov)
            s = D[I] - np.sum(L[I, 0:j] ** 2, axis=1)
            s[s < 0] = self.eps
            s = np.sqrt(s)
            tl = (self.lb[I] - L[I, 0:j] @ z[0:j]) / s
            tu = (self.ub[I] - L[I, 0:j] @ z[0:j]) / s
            pr[I] = lnNormalProb(tl, tu)
            # find smallest marginal dimension
            k = np.argmin(pr)

            # flip dimensions k-->j
            jk = [j, k]
            kj = [k, j]
            self.cov[jk, :] = self.cov[kj, :]  # update rows of cov
            self.cov[:, jk] = self.cov[:, kj]  # update cols of cov
            L[jk, :] = L[kj, :]  # update only rows of L
            self.lb[jk] = self.lb[kj]  # update integration limits
            self.ub[jk] = self.ub[kj]  # update integration limits
            perm[jk] = perm[kj]  # keep track of permutation

            # construct L sequentially via Cholesky computation
            s = self.cov[j, j] - np.sum(L[j, 0:j] ** 2, axis=0)
            if s < -0.01:
                raise RuntimeError("Sigma is not positive semi-definite")
            elif s < 0:
                s = self.eps
            L[j, j] = np.sqrt(s)
            new_L = self.cov[j + 1:self.dim, j] - L[j + 1:self.dim, 0:j] @ L[j, 0:j].T
            L[j + 1:self.dim, j] = new_L / L[j, j]

            # find mean value, z(j), of truncated normal
            tl = (self.lb[j] - L[j, 0:j - 1] @ z[0:j - 1]) / L[j, j]
            tu = (self.ub[j] - L[j, 0:j - 1] @ z[0:j - 1]) / L[j, j]
            w = lnNormalProb(tl, tu)  # aids in computing expected value of trunc. normal
            z[j] = (np.exp(-.5 * tl ** 2 - w) - np.exp(-.5 * tu ** 2 - w)) / np.sqrt(2 * math.pi)
        return L, perm


def lnNormalProb(a, b):
    # computes ln(P(a<Z<b)) where Z~N(0,1) very accurately for any 'a', 'b'
    p = np.zeros_like(a)
    # case b>a>0
    I = a > 0
    if np.any(I):
        pa = lnPhi(a[I])
        pb = lnPhi(b[I])
        p[I] = pa + np.log1p(-np.exp(pb - pa))
    # case a<b<0
    idx = b < 0
    if np.any(idx):
        pa = lnPhi(-a[idx])  # log of lower tail
        pb = lnPhi(-b[idx])
        p[idx] = pb + np.log1p(-np.exp(pa - pb))
    # case a < 0 < b
    I = (~I) & (~idx)
    if np.any(I):
        pa = special.erfc(-a[I] / np.sqrt(2)) / 2  # lower tail
        pb = special.erfc(b[I] / np.sqrt(2)) / 2  # upper tail
        p[I] = np.log1p(-pa - pb)
    return p


def lnPhi(x):
    # computes logarithm of  tail of Z~N(0,1) mitigating numerical roundoff errors
    out = -0.5 * x ** 2 - np.log(2) + np.log(special.erfcx(x / np.sqrt(2)) + EPS)  # divide by zeros error -> add eps
    return out

